/* -*- C++ -*- */
//=============================================================================
/**
 *  @file Virtual_Real_Time.h
 *
 *  $Id: Virtual_Real_Time.h,v 1.1.4.2 2002/01/21 22:35:06 phil Exp $
 *
 *  @author mesnier_p@ociweb.com
 */
  //===========================================================================


#ifndef VIRTUAL_REAL_TIME_H
#define VIRTUAL_REAL_TIME_H

#if defined (ACE_HAS_VIRTUAL_TIME)
# include <ace/OS.h>
# include <ace/Task.h>
# include <sys/times.h>
# include <limits.h>
#endif /* ACE_HAS_VIRTUAL_TIME */

class ACE_Thread_Manager;


class ACE_VRT_Synchronizer
{
 public:
  virtual void synchronize(void);
};

/**
 * @class ACE_Virtual_Real_Time
 *
 * @brief Controller of virtual time behavior
 *
 * The concept of virtual time is that a group of threads may behave as though 
 * operating in a time frame of reference other than the kernel's. This class
 * works in conjunction with the ACE_VRT_Wrapper class to adjust timeout values
 * depending on computations here. 
 *
 * The goal is to provide an environment on one platform that simuates the 
 * environment of another. Specfically, allowing certain threads in a Solaris
 * application to behave like VxWorks tasks.  Another goal is to allow two
 * or more processes to control the execution of code "in lock step" by 
 * suspending execution in one process until the others have caught up.
 * 
 * This goal is acheved by declaring one thread to be the controller and the
 * rest to be test threads. The controller operates in normal kernal time, and
 * is responsible for managing the test threads.
 *
 * The virtual time used by the test threads is a computed value. The value is
 * the current process time divided by the number of CPUs and multiplied by a
 * scale factor. 
 */

class ACE_Virtual_Real_Time 
#if !defined (ACE_HAS_VIRTUAL_TIME)
// make a null class
{
};
#else 
// do the real definition
: public ACE_Task<ACE_MT_SYNCH>
{

 public:
  ~ACE_Virtual_Real_Time (void);

  // To ensure that there is only a single instance of the virtual real time
  // object, the Singleton pattern is employed. Access to the singleton is 
  // through this public static method. If no instance is yet created, one will
  // be created in a thread safe manner.
  static ACE_Virtual_Real_Time *instance(void);

  // Get & set the thread manager used in conjuction with VRT. If no thread
  // manager is explicitly set, the singleton thread manager will be used.
  void test_thr_mgr (ACE_Thread_Manager *);
  ACE_Thread_Manager *test_thr_mgr (void);

  // Get & set the owner thread. Only the owner thread may control the test 
  // threads. By default the owner is the thread that creates the instance of
  // the VRT object, but if it is desired to run in a thread other than the
  // creator, that can be done by calling set_owner in the new thread. The
  // new owner may not be a member of the test thread group. 
  ACE_thread_t owner(void);
  void set_owner(void);
  void set_secondary(void);

  // Get & set the test thread group id. This may be any group other than one
  // the owner is in, if it is in a group. The test will not work properly if
  // the test group id is set after any test threads have been created, as
  // threads cannot change their group ID.
  void test_group (int );
  int test_group (void);

  // Test to see if the current thread is a member of the test group. Currently
  // this function returns true if the calling thread is not the owner thread.
  // The proper implementation of this function would be to test if the calling
  // thread is a member of the test group, but that involves a call into the
  // thread manager which is guarded by a mutex. Since the thread manager makes
  // time-based calls while holding the mutex, these calls deadlock if we try
  // to externally check for group membership.  This can be worked around by
  // making VRT a derivitive of ACE_Thread_Manager and use the un-locked 
  // protected method for determining group membership, but it remains to be
  // seen if that is truely necessary.
  int in_test_group (void);

  // The virtual time is computed based on the elapsed process time 
  // multiplied by the scale factor, so the higher the factor, the faster 
  // virtual time elapses.
  void scale_factor (double );
  double scale_factor (void);

  // The count of CPUs on the test host. This is used to further constrain the
  // current virtual time, but not in computing the scaled time values.
  void cpu_count (int );
  int cpu_count (void);

  // In order to approximate timeouts relative to the scale factor, the "real"
  // time value must be *divided* by the scale factor. For instance, if the
  // scale factor is 2, then 2 virtual seconds elapse in one real second, so
  // to wait for 1 virtual second, the actual delay must be 1/2 of a second.
  ACE_Time_Value scaled_time_value (const ACE_Time_Value & );

  /// Test thread control fucntions
  // Run_one_tick enables VRT mode, resumes all test threads, waits for 
  // the supplied number of system clock ticks (10 ms resolution), then
  // suspends the test threads.  It is assumed that the test threads start out
  // suspended. This method should only be called by the control thread.
  void run_one_tick (long ticks);

  // Resume_test_threads enables VRT mode, then resumes all test threads then
  // returns control to the caller. It is up to the caller to later suspend the
  // test threads at some future time. This function may only be called by the
  // owning thread.
  void resume_test_threads(void);

  // Suspend_test_threads suspends the test threads. This function may only be 
  // called by the owning thread. This does not affect the VRT enabled state.
  void suspend_test_threads(void);

  // Release_test_threads disables VRT mode and resumes suspended threads. The
  // test threads run in real time until normal completion. However, if a
  // thread is waiting in a scaled timeout and virtual time is ahead of real
  // time, the thread will continue to wait until real time has caught up.
  void release_test_threads(void);

  /// support functions
  // Retrieve the current process time based on calling getrusage. Currently
  // only the user time is used. It is probably better to sum the user & system
  // time. As this is the basis for computing virtual time, this value should
  // further be constrained by the number of CPUs, as the process time includes
  // time spent on all CPUs.
  ACE_Time_Value current_process_time(void);

  // Returns the scaled time elapsed since the process started. 
  ACE_Time_Value current_tick_time(void);

  // Returns the current virtual time, which is the scaled elapsed "tick" time
  // added to the real process start time.
  ACE_Time_Value current_virtual_time(void);

  // Wait_longer compares the current virtual time to the expected end time,
  // and if the CVT has not reached EET, then the passed time_to_wait pointer
  // is assigned the difference and the function returns non-zero.
  int wait_longer (ACE_Time_Value expected_end_time, 
                   ACE_Time_Value *time_to_wait);


  /// Control methods
  // Svc takes ownership of the control thread then loops forever checking for
  // incoming messages. If the current state is "paused" then svc waits
  // forever for a message. If the resulting message changes the state to 
  // "stop"  the loop is terminated and the test threads are released.
  // When the state is "run" then each iteration polls the message queue then
  // calls run_one_tick.
  int svc (void); 

  // Set the state to "run", this method puts a message on the queue to 
  void run_test (void);
  void stop_test (void);
  void pause_test (void);

  void synch_delay (const ACE_Time_Value &tv);

  void synchronizer (ACE_VRT_Synchronizer *);
  ACE_VRT_Synchronizer * synchronizer(void);


 protected:
  ACE_Virtual_Real_Time(void);
 private:
  ACE_Virtual_Real_Time(const ACE_Virtual_Real_Time&);
  ACE_Virtual_Real_Time& operator= (const ACE_Virtual_Real_Time&);

  static ACE_Virtual_Real_Time *vrt_;

  ACE_Time_Value real_start_time_;
  ACE_Time_Value suspended_time_;
  ACE_Time_Value idle_time_;
  double scale_factor_;
  int cpu_count_;
  clock_t start_clock_time_;

  int test_group_;
  ACE_Thread_Manager *test_thr_mgr_;

  ACE_thread_t owner_;
  ACE_thread_t secondary_;

  int vrt_enabled_;
  ACE_Time_Value synch_delay_;
  int delete_sync_;
  ACE_VRT_Synchronizer *synchronizer_;
};
#endif /* !ACE_HAS_VIRTUAL_TIME */


#if defined (__ACE_INLINE__)
#include "ace/Virtual_Real_Time.inl"
#endif /* __ACE_INLINE__ */

#endif /* VIRTUAL_REAL_TIME_H */






